//
// ram.v -- a preliminary main memory implementation using block RAM
//


module ram(clk, reset,
           en, wr, size, addr,
           data_in, data_out, wt);
    input clk;
    input reset;
    input en;
    input wr;
    input [1:0] size;
    input [25:0] addr;
    input [31:0] data_in;
    output reg [31:0] data_out;
    output wt;

  wire enable;
  wire [11:0] word_addr;

  reg [7:0] mem0[0:4095];
  reg [7:0] do0;
  reg [7:0] di0;
  reg wr0;

  reg [7:0] mem1[0:4095];
  reg [7:0] do1;
  reg [7:0] di1;
  reg wr1;

  reg [7:0] mem2[0:4095];
  reg [7:0] do2;
  reg [7:0] di2;
  reg wr2;

  reg [7:0] mem3[0:4095];
  reg [7:0] do3;
  reg [7:0] di3;
  reg wr3;

  reg state;

  assign enable = en & ~(|addr[25:14]);
  assign word_addr[11:0] = addr[13:2];

  always @(posedge clk) begin
    if (enable) begin
      if (wr & wr0) begin
        mem0[word_addr] <= di0;
      end
      do0 <= mem0[word_addr];
    end
  end

  always @(posedge clk) begin
    if (enable) begin
      if (wr & wr1) begin
        mem1[word_addr] <= di1;
      end
      do1 <= mem1[word_addr];
    end
  end

  always @(posedge clk) begin
    if (enable) begin
      if (wr & wr2) begin
        mem2[word_addr] <= di2;
      end
      do2 <= mem2[word_addr];
    end
  end

  always @(posedge clk) begin
    if (enable) begin
      if (wr & wr3) begin
        mem3[word_addr] <= di3;
      end
      do3 <= mem3[word_addr];
    end
  end

  // read
  always @(*) begin
    case (size[1:0])
      2'b00:
        // byte
        begin
          data_out[31:24] = 8'h00;
          data_out[23:16] = 8'h00;
          data_out[15: 8] = 8'h00;
          if (addr[1] == 0) begin
            if (addr[0] == 0) begin
              data_out[ 7: 0] = do0[7:0];
            end else begin
              data_out[ 7: 0] = do1[7:0];
            end
          end else begin
            if (addr[0] == 0) begin
              data_out[ 7: 0] = do2[7:0];
            end else begin
              data_out[ 7: 0] = do3[7:0];
            end
          end
        end
      2'b01:
        // halfword
        begin
          data_out[31:24] = 8'h00;
          data_out[23:16] = 8'h00;
          if (addr[1] == 0) begin
            data_out[15: 8] = do0[7:0];
            data_out[ 7: 0] = do1[7:0];
          end else begin
            data_out[15: 8] = do2[7:0];
            data_out[ 7: 0] = do3[7:0];
          end
        end
      2'b10:
        // word
        begin
          data_out[31:24] = do0[7:0];
          data_out[23:16] = do1[7:0];
          data_out[15: 8] = do2[7:0];
          data_out[ 7: 0] = do3[7:0];
        end
      default:
        // illegal
        begin
          data_out[31:24] = 8'hxx;
          data_out[23:16] = 8'hxx;
          data_out[15: 8] = 8'hxx;
          data_out[ 7: 0] = 8'hxx;
        end
    endcase
  end

  // write
  always @(*) begin
    case (size[1:0])
      2'b00:
        // byte
        begin
          di0[7:0] = data_in[ 7: 0];
          di1[7:0] = data_in[ 7: 0];
          di2[7:0] = data_in[ 7: 0];
          di3[7:0] = data_in[ 7: 0];
          wr0 = ~addr[1] & ~addr[0];
          wr1 = ~addr[1] &  addr[0];
          wr2 =  addr[1] & ~addr[0];
          wr3 =  addr[1] &  addr[0];
        end
      2'b01:
        // halfword
        begin
          di0[7:0] = data_in[15: 8];
          di1[7:0] = data_in[ 7: 0];
          di2[7:0] = data_in[15: 8];
          di3[7:0] = data_in[ 7: 0];
          wr0 = ~addr[1];
          wr1 = ~addr[1];
          wr2 =  addr[1];
          wr3 =  addr[1];
        end
      2'b10:
        // word
        begin
          di0[7:0] = data_in[31:24];
          di1[7:0] = data_in[23:16];
          di2[7:0] = data_in[15: 8];
          di3[7:0] = data_in[ 7: 0];
          wr0 = 1'b1;
          wr1 = 1'b1;
          wr2 = 1'b1;
          wr3 = 1'b1;
        end
      default:
        // illegal
        begin
          di0[7:0] = 8'hxx;
          di1[7:0] = 8'hxx;
          di2[7:0] = 8'hxx;
          di3[7:0] = 8'hxx;
          wr0 = 1'b0;
          wr1 = 1'b0;
          wr2 = 1'b0;
          wr3 = 1'b0;
        end
    endcase
  end

  // generate a single wait state for reads
  always @(posedge clk) begin
    if (reset == 1) begin
      state <= 1'b0;
    end else begin
      case (state)
        1'b0:
          begin
            if (enable == 1 && wr == 0) begin
              state <= 1'b1;
            end
          end
        1'b1:
          begin
            state <= 1'b0;
          end
      endcase
    end
  end

  assign wt = (en == 1 && wr == 0 && state == 1'b0) ? 1 : 0;

endmodule
